/*
/*******************************************************************************
 * Copyright (C) 2024 Secretaría General de la Administración Digital, Gobierno de España
 * This program is licensed and may be used, modified and redistributed under the  terms
 * of the European Public License (EUPL), either version 1.1 or (at your option)
 * any later version as soon as they are approved by the European Commission.
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and
 * more details.
 * You should have received a copy of the EUPL1.1 license
 * along with this program; if not, you may find it at
 * http:joinup.ec.europa.eu/software/page/eupl/licence-eupl
 ******************************************************************************/

/**
 * <b>File:</b><p>es.clave.SP2.LogoutRequestController.java.</p>
 * <b>Description:</b><p>Class to generate a SAML request of type Logout. This request is sent to Proxy2 and is used to delete the session that the Kit has previously created in Proxy2.</p>
 * <b>Project:</b><p>Citizen identification and authentication platform.</p>
 * <b>Date:</b><p>03/10/2024.</p>
 * @author Gobierno de España.
 * @version 1.0, 03/10/2024.
 */
package es.clave.SP2;

import java.io.FileInputStream;
import java.security.KeyStore;
import java.security.KeyStore.PrivateKeyEntry;
import java.security.PrivateKey;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.time.Instant;
import java.util.Base64;

import org.opensaml.core.config.InitializationService;
import org.opensaml.core.xml.Namespace;
import org.opensaml.core.xml.config.XMLObjectProviderRegistrySupport;
import org.opensaml.core.xml.io.Marshaller;
import org.opensaml.saml.common.SAMLVersion;
import org.opensaml.saml.saml2.core.Issuer;
import org.opensaml.saml.saml2.core.LogoutRequest;
import org.opensaml.saml.saml2.core.NameID;
import org.opensaml.saml.saml2.core.impl.IssuerBuilder;
import org.opensaml.saml.saml2.core.impl.LogoutRequestBuilder;
import org.opensaml.saml.saml2.core.impl.NameIDBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PostMapping;
import org.w3c.dom.Element;

import es.clave.SP2.util.Utils;
import net.shibboleth.utilities.java.support.xml.SerializeSupport;

/**
 * <p>Class to generate a SAML request of type Logout. This request is sent to Proxy2 and is used to delete the session that the Kit has previously created in Proxy2.</p>
 * <b>Project:</b><p>Citizen identification and authentication platform.</p>
 * @version 1.0, 03/10/2024.
 */
@Controller
public class LogoutRequestController {

	private static final Logger logger = LoggerFactory.getLogger(LogoutRequestController.class);

	/**
	 * Injecting the value 'request.providerName' from the application.properties file
	 */
	@Value("${request.providerName}")
	private String providerName;

	/**
	 * Injecting the value 'request.url' from the application.properties file
	 */
	@Value("${request.url}")
	private String requestUrl;

	/**
	 * Injecting the value 'request.returnURL' from the application.properties file
	 */
	@Value("${request.returnURL}")
	private String returnURL;

	/**
	 * Injecting the value 'sign.keyStorePath' from the application.properties file
	 */
	@Value("${sign.keyStorePath}")
	private String keyStorePath;

	/**
	 * Injecting the value 'sign.keyStorePassword' from the application.properties file
	 */
	@Value("${sign.keyStorePassword}")
	private String keyStorePassword;

	/**
	 * Injecting the value 'sign.keyPassword' from the application.properties file
	 */
	@Value("${sign.keyPassword}")
	private String keyPassword;

	/**
	 * Injecting the value 'sign.alias' from the application.properties file
	 */
	@Value("${sign.alias}")
	private String alias;

	/**
	 * Injecting the value 'signature.algorithm' from the application.properties file
	 */
	@Value("${signature.algorithm}")
	private String signatureAlgorithm;

	/**
	 * Function to generate a SAML request of type Logout. This request is sent to Proxy2 and is used to delete the session that the Kit has previously created in Proxy2
	 *
	 * @param requestSAML
	 * @param model
	 * @return The name of the view to invoke
	 */
	@PostMapping(value = "/logout")
	public String logoutRequest(@ModelAttribute RequestSAML requestSAML, Model model) {
		model.addAttribute("request", requestSAML);
		logger.info("logoutRequest" + requestSAML);

		try {
			// Se inicializa la configuración de OpenSaml
			InitializationService.initialize();
			// Se instancian los objetos que almacenarán la información con la
			// que construir la petición SAML
			LogoutRequest logoutRequest = new LogoutRequestBuilder().buildObject();

			// Define la URL de respuesta con la que deberá de comunicarse
			// Pasarela (Proxy2) cuando finalice el ciclo de desconexión.
			Issuer issuer = new IssuerBuilder().buildObject();
			issuer.setValue(returnURL);

			// Se inicializa el atributo que identifica el nameIDPolicy.
			NameID nameID = new NameIDBuilder().buildObject();
			nameID.setFormat(NameID.UNSPECIFIED);
			nameID.setSPNameQualifier(returnURL);
			nameID.setValue(providerName);

			// Definición de namespaces por defecto asociados al xml (SAML)
			logoutRequest.getNamespaceManager().registerNamespaceDeclaration(new Namespace("urn:oasis:names:tc:SAML:2.0:assertion", "saml2"));
			logoutRequest.getNamespaceManager().registerNamespaceDeclaration(new Namespace("http://www.w3.org/2000/09/xmldsig#", "ds"));
			logoutRequest.getNamespaceManager().registerNamespaceDeclaration(new Namespace("urn:oasis:names:tc:SAML:2.0:protocol", "saml2p"));
			logoutRequest.getNamespaceManager().registerNamespaceDeclaration(new Namespace("http://eidas.europa.eu/saml-extensions", "eidas"));

			// A continuación se establecen los valores propios de la petición
			// SAML de desconexión. Para más información sobre los atributos:
			// https://docs.oasis-open.org/security/saml/v2.0/saml-core-2.0-os.pdf
			logoutRequest.setIssueInstant(Instant.now());
			logoutRequest.setDestination(requestUrl);
			logoutRequest.setID(SecureRandomXmlIdGenerator.INSTANCE.generateIdentifier());
			logoutRequest.setIssuer(issuer);
			logoutRequest.setNameID(nameID);
			logoutRequest.setVersion(SAMLVersion.VERSION_20);
			logoutRequest.setReason("urn:oasis:names:tc:SAML:2.0:logout:user");
			logoutRequest.setNotOnOrAfter(Instant.now().plusSeconds(300));

			// Se inicializa la configuración de la firma de la petición SAML de
			// desconexión a partir de los valores establecidos en el fichero
			// application.properties
			KeyStore ks = KeyStore.getInstance(KeyStore.getDefaultType());
			try (FileInputStream fis = new FileInputStream(keyStorePath)) {
				ks.load(fis, keyStorePassword.toCharArray());
			}
			KeyStore.PrivateKeyEntry pkEntry = (PrivateKeyEntry) ks.getEntry(alias, new KeyStore.PasswordProtection(keyPassword.toCharArray()));
			PrivateKey pk = pkEntry.getPrivateKey();
			Certificate certificate = pkEntry.getCertificate();

			// Se realiza la firma de la petición SAML de desconexión construida
			Utils.signRequest(logoutRequest, pk, (X509Certificate) certificate, signatureAlgorithm);
			Marshaller out = XMLObjectProviderRegistrySupport.getMarshallerFactory().getMarshaller(logoutRequest);
			out.marshall(logoutRequest);

			// Se lleva a cabo la lógica para convertir la petición SAML de
			// desconexión construida con OpenSaml en un documento XML para
			// codificarlo posteriormente en Base64 como lo espera recibir
			// Pasarela (Proxy2)
			Element elementDOM = logoutRequest.getDOM();
			String xmlString = SerializeSupport.nodeToString(elementDOM);

			String encodedString = Base64.getEncoder().encodeToString(xmlString.getBytes());

			// Se genera el identificador RelayState que permitirá su validación
			// en la respuesta desde Pasarela (Proxy2)
			String relayState = SecureRandomXmlIdGenerator.INSTANCE.generateIdentifier(8);
			requestSAML.setRelayState(relayState);
			requestSAML.setLogoutRequest(encodedString);
			requestSAML.setUrl(requestUrl);

			SessionHolder.sessionsSAML.put(logoutRequest.getID(), relayState);

			logger.info("Petición de Logout con RelayState: " + requestSAML.getRelayState() + " y LogoutRequest: " + requestSAML.getLogoutRequest());

		} catch (Exception e) {
			String errorMessage = "Ha habido un error en la contrucción de la petición LOGOUT.";
			logger.error(errorMessage, e);
			model.addAttribute("errorMessageErrorPage", errorMessage);
			return "error";
		}

		return "redirectLogoutRequest";
	}
}
