/* 
/*******************************************************************************
 * Copyright (C) 2024 Secretaría General de la Administración Digital, Gobierno de España
 * This program is licensed and may be used, modified and redistributed under the  terms
 * of the European Public License (EUPL), either version 1.1 or (at your option)
 * any later version as soon as they are approved by the European Commission.
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and
 * more details.
 * You should have received a copy of the EUPL1.1 license
 * along with this program; if not, you may find it at
 * http:joinup.ec.europa.eu/software/page/eupl/licence-eupl
 ******************************************************************************/

/** 
 * <b>File:</b><p>es.clave.SP2.ResponseController.java.</p>
 * <b>Description:</b><p>Class that processes the SAML response that Proxy2 sends to the Kit.</p>
 * <b>Project:</b><p>Citizen identification and authentication platform.</p>
 * <b>Date:</b><p>03/10/2024.</p>
 * @author Gobierno de España.
 * @version 1.0, 03/10/2024.
 */
package es.clave.SP2;

import java.io.ByteArrayInputStream;
import java.util.Base64;
import java.util.HashMap;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.opensaml.core.config.InitializationService;
import org.opensaml.core.xml.XMLObject;
import org.opensaml.core.xml.config.XMLObjectProviderRegistrySupport;
import org.opensaml.core.xml.io.Unmarshaller;
import org.opensaml.core.xml.io.UnmarshallerFactory;
import org.opensaml.saml.saml2.core.Assertion;
import org.opensaml.saml.saml2.core.LogoutResponse;
import org.opensaml.saml.saml2.core.Response;
import org.opensaml.saml.saml2.core.StatusCode;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PostMapping;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/** 
 * <p>Class that processes the SAML response that Proxy2 sends to the Kit.</p>
 * <b>Project:</b><p>Citizen identification and authentication platform.</p>
 * @version 1.0, 03/10/2024.
 */
@Controller
public class ResponseController {

	private static final Logger logger = LoggerFactory.getLogger(ResponseController.class);

	/**
	 * Function that processes the SAML response that Proxy2 sends to the Kit.
	 * 
	 * @param response
	 * @param model
	 * @return Class that processes the SAML response that Proxy2 sends to the Kit.
	 * @throws Exception
	 */
	@PostMapping("/ReturnPage")
	public String returnAction(@ModelAttribute ResponseSAML response, Model model) throws Exception {
		model.addAttribute("response", new ResponseSAML());

		logger.info("response.relayState: " + response.getRelayState());
		logger.info("response.SAMLResponse: " + response.getSAMLResponse());
		logger.info("response.logoutResponse: " + response.getLogoutResponse());

		// En este método se lleva a cabo la validación de la respuesta SAML
		// tanto para la solicitud de autenticación como la de desconexión.
		// Se comprueba si la respuesta recibida desde Pasarela corresponde a
		// una respuesta de autenticación (Response)
		if (response.getSAMLResponse() != null && !response.getSAMLResponse().trim().isEmpty()) {
			// Se inicializa la configuración de OpenSaml
			InitializationService.initialize();

			// Se decodifica la respuesta SAML recibida en Base64
			byte[ ] decodedBytes = Base64.getDecoder().decode(response.getSAMLResponse());

			// Se transforma respuesta SAML en un documento XML
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			factory.setFeature("http://apache.org/xml/features/disallow-doctype-decl", true);
			factory.setNamespaceAware(true);
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document document = builder.parse(new ByteArrayInputStream(decodedBytes));
			Element element = document.getDocumentElement();

			// Se lleva a cabo la conversión del documento XML en el objeto SAML
			// Response
			UnmarshallerFactory unmarshallerFactory = XMLObjectProviderRegistrySupport.getUnmarshallerFactory();
			Unmarshaller unmarshaller = unmarshallerFactory.getUnmarshaller(element);
			XMLObject xmlObject = unmarshaller.unmarshall(element);

			if (xmlObject instanceof Response) {
				Response samlResponse = (Response) xmlObject;
				// Se obtiene la información del RelayState desde la respuesta
				// SAML para verificar que coincide con el que generó durante el
				// envío para así verificar su integridad.
				String prevRelayState = SessionHolder.sessionsSAML.get(samlResponse.getInResponseTo());
				SessionHolder.sessionsSAML.clear();
				// Si no coincide el RelayState se detiene el proceso y se
				// muestra un mensaje de error al usuario.
				if (prevRelayState == null || !prevRelayState.equals(response.getRelayState())) {
					String errorMessage = "Ha habido un error de verificación del RelayState para la respuesta de tipo samlResponse recibida.";
					logger.error(errorMessage);
					model.addAttribute("errorMessageErrorPage", errorMessage);
					return "error";
				}

				// Validar si la respuesta de autenticación desde Pasarela ha resultado exitosa
				if (samlResponse.getStatus().getStatusCode().getValue().equals(StatusCode.SUCCESS)) {
					// Autenticación correcta
    				// Se recorre el objeto Response para acceder a la información
    				// recibida desde Pasarela (Proxy2)
    				for (Assertion assertion: samlResponse.getAssertions()) {
    					logger.info("Issuer: " + assertion.getIssuer().getValue());
    					logger.info("Subject: " + assertion.getSubject().getNameID().getValue());
    					Map<String, String> map = new HashMap<>();
    
    					// Se obtienen cada uno de los atributos existentes en el
    					// SAML y, en este caso, se almacenan en una variable que se
    					// utilizará para mostrar la información al usuario en
    					// pantalla.
    					assertion.getAttributeStatements().forEach(attrStatement -> {
    						attrStatement.getAttributes().forEach(attribute -> {
    							logger.info("Attribute Name: " + attribute.getFriendlyName());
    							attribute.getAttributeValues().forEach(value -> {
    								logger.info("Attribute Value: " + value.getDOM().getTextContent());
    								map.put(attribute.getFriendlyName(), value.getDOM().getTextContent());
    							});
    						});
    					});
    
    					model.addAttribute("map", map);
    				}
				} else {
					String errorMessage = "Se ha producido un error no controlado durante la petición de autenticación";
					if (samlResponse.getStatus() != null && samlResponse.getStatus().getStatusMessage() != null) {
						errorMessage = samlResponse.getStatus().getStatusMessage().getValue();
					}
					logger.error("Error de autenticación devuelto por Pasarela: " + errorMessage);
					model.addAttribute("errorMessageErrorPage", errorMessage);
					return "error";
				}

				return "returnPage";

			} else {
				String errorMessage = "Respuesta de tipo samlResponse inválida.";
				logger.error(errorMessage);
				model.addAttribute("errorMessageErrorPage", errorMessage);
				return "error";
			}

			// Se comprueba si la respuesta recibida desde Pasarela corresponde
			// a una respuesta de solicitud de desconexión (LogoutResponse)
		} else if (response.getLogoutResponse() != null && !response.getLogoutResponse().trim().isEmpty()) {
			// Se inicializa la configuración de OpenSaml
			InitializationService.initialize();

			// Se decodifica la respuesta SAML recibida en Base64
			byte[ ] decodedBytes = Base64.getDecoder().decode(response.getLogoutResponse());

			// Se transforma respuesta SAML en un documento XML
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			factory.setFeature("http://apache.org/xml/features/disallow-doctype-decl", true);
			factory.setNamespaceAware(true);
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document document = builder.parse(new ByteArrayInputStream(decodedBytes));
			Element element = document.getDocumentElement();

			// Se lleva a cabo la conversión del documento XML en el objeto SAML
			// LogoutResponse
			UnmarshallerFactory unmarshallerFactory = XMLObjectProviderRegistrySupport.getUnmarshallerFactory();
			Unmarshaller unmarshaller = unmarshallerFactory.getUnmarshaller(element);
			XMLObject xmlObject = unmarshaller.unmarshall(element);

			if (xmlObject instanceof LogoutResponse) {
				LogoutResponse logoutResponse = (LogoutResponse) xmlObject;
				// Se obtiene la información del RelayState desde la respuesta
				// SAML para verificar que coincide con el que generó durante el
				// envío para así verificar su integridad.
				String prevRelayState = SessionHolder.sessionsSAML.get(logoutResponse.getInResponseTo());
				SessionHolder.sessionsSAML.clear();
				// Si no coincide el RelayState se detiene el proceso y se
				// muestra un mensaje de error al usuario.
				if (prevRelayState == null || !prevRelayState.equals(response.getRelayState())) {
					String errorMessage = "Ha habido un error de verificación del RelayState para la respuesta de tipo logoutResponse recibida.";
					logger.error(errorMessage);
					model.addAttribute("errorMessageErrorPage", errorMessage);
					return "error";
				}
				// Se redirecciona al usuario a la pantalla principal del kit ya
				// que el proceso de desconexión se ha realizado correctamente.
				return "redirect:/IndexPage";

			} else {
				String errorMessage = "Respuesta de tipo logoutResponse inválida.";
				logger.error(errorMessage);
				model.addAttribute("errorMessageErrorPage", errorMessage);
				return "error";
			}

		} else {
			String errorMessage = "Se ha recibido una respuesta vacía.";
			logger.error(errorMessage);
			model.addAttribute("errorMessageErrorPage", errorMessage);
			return "error";
		}
	}
}
