/*
/*******************************************************************************
 * Copyright (C) 2024 Secretaría General de la Administración Digital, Gobierno de España
 * This program is licensed and may be used, modified and redistributed under the  terms
 * of the European Public License (EUPL), either version 1.1 or (at your option)
 * any later version as soon as they are approved by the European Commission.
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and
 * more details.
 * You should have received a copy of the EUPL1.1 license
 * along with this program; if not, you may find it at
 * http:joinup.ec.europa.eu/software/page/eupl/licence-eupl
 ******************************************************************************/

/**
 * <b>File:</b><p>es.clave.SP2.SubmitRequestController.java.</p>
 * <b>Description:</b><p>Class to build the SAML request that is sent from the KIT to PROXY2.</p>
 * <b>Project:</b><p>Citizen identification and authentication platform.</p>
 * <b>Date:</b><p>03/10/2024.</p>
 * @author Gobierno de España.
 * @version 1.0, 03/10/2024.
 */
package es.clave.SP2;

import java.io.FileInputStream;
import java.security.KeyStore;
import java.security.KeyStore.PrivateKeyEntry;
import java.security.PrivateKey;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Base64;
import java.util.List;

import javax.xml.namespace.QName;

import org.opensaml.core.config.InitializationService;
import org.opensaml.core.xml.Namespace;
import org.opensaml.core.xml.XMLObjectBuilderFactory;
import org.opensaml.core.xml.config.XMLObjectProviderRegistrySupport;
import org.opensaml.core.xml.io.Marshaller;
import org.opensaml.core.xml.io.MarshallingException;
import org.opensaml.core.xml.schema.XSAny;
import org.opensaml.core.xml.schema.impl.XSAnyBuilder;
import org.opensaml.saml.common.SAMLVersion;
import org.opensaml.saml.saml2.core.AuthnContextClassRef;
import org.opensaml.saml.saml2.core.AuthnContextComparisonTypeEnumeration;
import org.opensaml.saml.saml2.core.AuthnRequest;
import org.opensaml.saml.saml2.core.Extensions;
import org.opensaml.saml.saml2.core.NameIDPolicy;
import org.opensaml.saml.saml2.core.RequestAbstractType;
import org.opensaml.saml.saml2.core.RequestedAuthnContext;
import org.opensaml.saml.saml2.core.impl.AuthnContextClassRefBuilder;
import org.opensaml.saml.saml2.core.impl.AuthnRequestBuilder;
import org.opensaml.saml.saml2.core.impl.NameIDPolicyBuilder;
import org.opensaml.saml.saml2.core.impl.RequestedAuthnContextBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PostMapping;
import org.w3c.dom.Element;

import es.clave.SP2.util.Utils;
import net.shibboleth.utilities.java.support.xml.SerializeSupport;

/**
 * <p>Class to build the SAML request that is sent from the KIT to PROXY2.</p>
 * <b>Project:</b><p>Citizen identification and authentication platform.</p>
 * @version 1.0, 03/10/2024.
 */
@Controller
public class SubmitRequestController {

	private static final Logger logger = LoggerFactory.getLogger(SubmitRequestController.class);

	/**
	 * Injecting the value 'sign.keyStorePath' from the application.properties file
	 */
	@Value("${sign.keyStorePath}")
	private String keyStorePath;

	/**
	 * Injecting the value 'sign.keyStorePassword' from the application.properties file
	 */
	@Value("${sign.keyStorePassword}")
	private String keyStorePassword;

	/**
	 * Injecting the value 'sign.keyPassword' from the application.properties file
	 */
	@Value("${sign.keyPassword}")
	private String keyPassword;

	/**
	 * Injecting the value 'sign.alias' from the application.properties file
	 */
	@Value("${sign.alias}")
	private String alias;

	/**
	 * Injecting the value 'signature.algorithm' from the application.properties file
	 */
	@Value("${signature.algorithm}")
	private String signatureAlgorithm;

	/**
	 * Function that builds the SAML request that is sent from KIT to PROXY2
	 *
	 * @param requestSAML
	 * @param model
	 * @return The name of the view to invoke
	 */
	@PostMapping(value = "/getParams")
	public String submitRequest(@ModelAttribute RequestSAML requestSAML, Model model) {
		model.addAttribute("request", requestSAML);
		logger.info("submitRequest" + requestSAML);

		try {
			// Se inicializa la configuración de OpenSaml
			InitializationService.initialize();
			// Se instancian los objetos que almacenarán la información con la
			// que construir la petición SAML
			AuthnRequest authnRequest = new AuthnRequestBuilder().buildObject();
			NameIDPolicy nameIDPolicy = new NameIDPolicyBuilder().buildObject();
			// Se inicializa el atributo que identifica el nameIDPolicy.
			nameIDPolicy.setAllowCreate(true);
			nameIDPolicy.setFormat(requestSAML.getNameIDPolicy());

			// Definición de namespaces por defecto asociados al xml (SAML)
			authnRequest.getNamespaceManager().registerNamespaceDeclaration(new Namespace("urn:oasis:names:tc:SAML:2.0:assertion", "saml2"));
			authnRequest.getNamespaceManager().registerNamespaceDeclaration(new Namespace("http://www.w3.org/2000/09/xmldsig#", "ds"));
			authnRequest.getNamespaceManager().registerNamespaceDeclaration(new Namespace("urn:oasis:names:tc:SAML:2.0:protocol", "saml2p"));
			authnRequest.getNamespaceManager().registerNamespaceDeclaration(new Namespace("http://eidas.europa.eu/saml-extensions", "eidas"));
			authnRequest.getNamespaceManager().registerNamespaceDeclaration(new Namespace("http://eidas.europa.eu/attributes/naturalperson", "eidas-natural"));

			// A continuación se establecen los valores propios de la petición
			// SAML. Para más información sobre los atributos:
			// https://docs.oasis-open.org/security/saml/v2.0/saml-core-2.0-os.pdf
			// Atributo opcional que informa al receptor que el usuario permite
			// el consentimiento del envío de la petición SAML. Por defecto:
			// Unspecified.
			authnRequest.setConsent(RequestAbstractType.UNSPECIFIED_CONSENT);
			// Define la URL de respuesta con la que deberá de comunicarse
			// Pasarela (Proxy2) cuando finalice el ciclo de autenticación.
			authnRequest.setAssertionConsumerServiceURL(requestSAML.getReturnURL());
			// Define la URL del entorno de Pasarela (Proxy2) a la que se está
			// enviando la petición SAML.
			authnRequest.setDestination(requestSAML.getUrl());
			// Define si se quiere forzar una nueva autenticación en Pasarela
			// (Proxy2).
			authnRequest.setForceAuthn(requestSAML.getForceCheck());
			// Número aleatorio que identifica la petición SAML.
			authnRequest.setID(SecureRandomXmlIdGenerator.INSTANCE.generateIdentifier());
			authnRequest.setIsPassive(false);
			authnRequest.setIssueInstant(Instant.now());
			// Define la información del SP que realiza la petición de
			// autenticación en Pasarela (Proxy2)
			authnRequest.setProviderName(requestSAML.getProviderName() + ";" + requestSAML.getApplication());
			authnRequest.setVersion(SAMLVersion.VERSION_20);
			authnRequest.setNameIDPolicy(nameIDPolicy);

			// Define el atributo que establece el contexto de la autenticación.
			RequestedAuthnContext authContext = new RequestedAuthnContextBuilder().buildObject();
			authContext.setComparison(AuthnContextComparisonTypeEnumeration.MINIMUM);
			AuthnContextClassRefBuilder authnContextClassRefBuilder = new AuthnContextClassRefBuilder();
			AuthnContextClassRef authnContextClassRef = authnContextClassRefBuilder.buildObject();
			// Define el nivel de seguridad que aplica en esta petición SAML
			// (Level of Assurance (LoA))
			authnContextClassRef.setURI(requestSAML.getEidasloa());
			authContext.getAuthnContextClassRefs().add(authnContextClassRef);

			// Define los atributos que se enviaran en la petición SAML y que,
			// en este caso, permitirán deshabilitar determinados IdPs en
			// Pasarela (Proxy2)
			List<EidasAttribute> requestEidasAttributes = new ArrayList<EidasAttribute>();
			if (requestSAML.getAfirmaCheck()) {
				requestEidasAttributes.add(EidasAttribute.AFIRMA);
			}
			if (requestSAML.getGissCheck()) {
				requestEidasAttributes.add(EidasAttribute.GISS);
			}
			if (requestSAML.getAeatCheck()) {
				requestEidasAttributes.add(EidasAttribute.AEAT);
			}
			if (requestSAML.getEidasCheck()) {
				requestEidasAttributes.add(EidasAttribute.EIDAS);
			}
			if (requestSAML.getMobileCheck()) {
				requestEidasAttributes.add(EidasAttribute.MOVIL);
			}

			requestEidasAttributes.add(EidasAttribute.RELAY_STATE);
			authnRequest.setExtensions(buildExtensions(requestEidasAttributes));
			authnRequest.setRequestedAuthnContext(authContext);

			// Se inicializa la configuración de la firma de la petición SAML a
			// partir de los valores establecidos en el fichero
			// application.properties
			KeyStore ks = KeyStore.getInstance(KeyStore.getDefaultType());
			try (FileInputStream fis = new FileInputStream(keyStorePath)) {
				ks.load(fis, keyStorePassword.toCharArray());
			}
			KeyStore.PrivateKeyEntry pkEntry = (PrivateKeyEntry) ks.getEntry(alias, new KeyStore.PasswordProtection(keyPassword.toCharArray()));
			PrivateKey pk = pkEntry.getPrivateKey();
			Certificate certificate = pkEntry.getCertificate();

			// Se realiza la firma de la petición SAML construida
			Utils.signRequest(authnRequest, pk, (X509Certificate) certificate, signatureAlgorithm);
			Marshaller out = XMLObjectProviderRegistrySupport.getMarshallerFactory().getMarshaller(authnRequest);
			out.marshall(authnRequest);

			// Se lleva a cabo la lógica para convertir la petición SAML
			// construida con OpenSaml en un documento XML para codificarlo
			// posteriormente en Base64 como lo espera recibir Pasarela (Proxy2)
			Element elementDOM = authnRequest.getDOM();
			String xmlString = SerializeSupport.nodeToString(elementDOM);

			String encodedString = Base64.getEncoder().encodeToString(xmlString.getBytes());

			// Se genera el identificador RelayState que permitirá su validación
			// en la respuesta desde Pasarela (Proxy2)
			String relayState = SecureRandomXmlIdGenerator.INSTANCE.generateIdentifier(8);
			requestSAML.setRelayState(relayState);
			requestSAML.setSAMLRequest(encodedString);

			SessionHolder.sessionsSAML.put(authnRequest.getID(), relayState);
			logger.info("Petición con RelayState: " + requestSAML.getRelayState() + " y SAMLRequest: " + requestSAML.getSAMLRequest());

		} catch (Exception e) {
			String errorMessage = "Ha habido un error en la contrucción de la petición SAML.";
			logger.error(errorMessage, e);
			model.addAttribute("errorMessageErrorPage", errorMessage);
			return "error";
		}

		return "redirectSubmitRequest";
	}

	private Extensions buildExtensions(List<EidasAttribute> eidasAttributes) throws Exception {
		Extensions extensions = buildSAMLObject(Extensions.class);

		XSAny requestedAttributes = new XSAnyBuilder().buildObject("http://eidas.europa.eu/saml-extensions", "RequestedAttributes", "eidas");
		addEidasAttributes(eidasAttributes, requestedAttributes);
		extensions.getUnknownXMLObjects().add(requestedAttributes);

		return extensions;
	}

	public <T> T buildSAMLObject(final Class<T> clazz) throws Exception {
		T object = null;

		try {
			XMLObjectBuilderFactory builderFactory = XMLObjectProviderRegistrySupport.getBuilderFactory();
			QName defaultElementName = (QName) clazz.getDeclaredField("DEFAULT_ELEMENT_NAME").get(null);
			Object rawObject = builderFactory.getBuilder(defaultElementName).buildObject(defaultElementName);
			if (clazz.isInstance(rawObject)) {
				object = clazz.cast(rawObject);
			} else {
				throw new ClassCastException("Failed to cast the SAML object to the expected type " + clazz.getName());
			}

		} catch (IllegalAccessException | NoSuchFieldException e) {
			logger.error("buildSAMLObject - Error al crear el SAML object", e);
			throw e;
		}

		return object;
	}

	private void addEidasAttributes(List<EidasAttribute> eidasAttributes, XSAny requestedAttributes) throws Exception {
		if (eidasAttributes == null) {
			return;
		}
		for (EidasAttribute attribute: eidasAttributes) {
			addEidasAttribute(requestedAttributes, attribute);
		}
	}

	private void addEidasAttribute(XSAny requestedAttributes, EidasAttribute eidasAttribute) throws Exception {
		try {
			requestedAttributes.getUnknownXMLObjects().add(buildRequestedAttribute(eidasAttribute));
		} catch (MarshallingException e) {
			logger.error("addEidasAttribute - Error al realizar el marshalling.", e);
			throw e;
		}
	}

	private XSAny buildRequestedAttribute(EidasAttribute eidasAttribute) throws MarshallingException {
		XSAny requestedAttribute = new XSAnyBuilder().buildObject("http://eidas.europa.eu/saml-extensions", "RequestedAttribute", "eidas");
		requestedAttribute.getUnknownAttributes().put(new QName("FriendlyName"), eidasAttribute.getFriendlyName());
		requestedAttribute.getUnknownAttributes().put(new QName("Name"), eidasAttribute.getName());
		requestedAttribute.getUnknownAttributes().put(new QName("NameFormat"), "urn:oasis:names:tc:SAML:2.0:attrname-format:uri");
		requestedAttribute.getUnknownAttributes().put(new QName("isRequired"), eidasAttribute.isRequired() ? "true" : "false");

		if (eidasAttribute.equals(EidasAttribute.RELAY_STATE)) {
			XSAny attributeValue = new XSAnyBuilder().buildObject("http://eidas.europa.eu/saml-extensions", "AttributeValue", "eidas");
			attributeValue.setTextContent(SecureRandomXmlIdGenerator.INSTANCE.generateIdentifier(8));
			attributeValue.getUnknownAttributes().put(new QName("http://www.w3.org/2001/XMLSchema-instance", "type", "xsi"), "eidas-natural:PersonIdentifierType");

			requestedAttribute.getUnknownXMLObjects().add(attributeValue);
		}

		return requestedAttribute;
	}
}
