/*
/*******************************************************************************
 * Copyright (C) 2024 Secretaría General de la Administración Digital, Gobierno de España
 * This program is licensed and may be used, modified and redistributed under the  terms
 * of the European Public License (EUPL), either version 1.1 or (at your option)
 * any later version as soon as they are approved by the European Commission.
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and
 * more details.
 * You should have received a copy of the EUPL1.1 license
 * along with this program; if not, you may find it at
 * http:joinup.ec.europa.eu/software/page/eupl/licence-eupl
 ******************************************************************************/

/**
 * <b>File:</b><p>es.clave.SP2.daemon.DaemonCertificateProxy2.java.</p>
 * <b>Description:</b><p>Class that periodically executes a Daemon-type process, which makes a call to a Proxy2 servlet to download certificates with private keys, both for signature and encryption. The process of obtaining them is done through an XML.</p>
 * <b>Project:</b><p>Citizen identification and authentication platform.</p>
 * <b>Date:</b><p>03/10/2024.</p>
 * @author Gobierno de España.
 * @version 1.0, 03/10/2024.
 */
package es.clave.SP2.daemon;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.InetAddress;
import java.net.URL;
import java.net.UnknownHostException;
import java.security.KeyStore;
import java.security.SecureRandom;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Base64;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManagerFactory;
import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import es.clave.SP2.util.MessagesPropertiesRetriever;
import es.clave.SP2.util.Utils;

/**
 * <p>Class that periodically executes a Daemon-type process, which makes a call to a Proxy2 servlet to download certificates with private keys, both for signature and encryption. The process of obtaining them is done through an XML.</p>
 * <b>Project:</b><p>Citizen identification and authentication platform.</p>
 * @version 1.0, 03/10/2024.
 */
@Component
public class DaemonCertificateProxy2 {

	private static final Logger logger = LoggerFactory.getLogger(DaemonCertificateProxy2.class);

	private MessagesPropertiesRetriever messagesProperties;

	/**
	 * Injecting the value 'certProxy2.daemonActivated' from the application.properties file
	 */
	@Value("${certProxy2.daemonActivated}")
	private String daemonActivated;

	/**
	 * Injecting the value 'certproxy2.endpoint' from the application.properties file
	 */
	@Value("${certproxy2.endpoint}")
	private String endpoint;

	/**
	 * Injecting the value 'certproxy2.truststorePath' from the application.properties file
	 */
	@Value("${certproxy2.truststorePath}")
	private String truststorePath;

	/**
	 * Injecting the value 'certProxy2.truststorePassword' from the application.properties file
	 */
	@Value("${certProxy2.truststorePassword}")
	private String truststorePassword;

	/**
	 * Injecting the value 'certProxy2.certificatesPath' from the application.properties file
	 */
	@Value("${certProxy2.certificatesPath}")
	private String certificatesPath;

	/**
	 * Constructor method for the class DaemonCertificateProxy2.java.
	 * @param messagesProperties The MessagesPropertiesRetriever class is injected through the constructor
	 */
	public DaemonCertificateProxy2(MessagesPropertiesRetriever messagesProperties) {
		this.messagesProperties = messagesProperties;
	}

	/**
	 * Daemon-type process, which makes a call to a Proxy2 servlet to download certificates with private keys, both for signature and encryption. The process of obtaining them is done through an XML.
	 * The daemon runs every X time. This time is configured in milliseconds in the variable 'certProxy2.daemonFixedRate' in the application.properties file
	 */
	@Scheduled(fixedRateString = "${certProxy2.daemonFixedRate}")
	public void recoverAndUpdatePublicCertificates() {
		String hostname = null;
		String message;
		
		// Comprobar si existe directorio configurado en la propiedad "certProxy2.certificatesPath" del fichero application.properties. En caso de que no exista se utilizará la ruta por defecto establecida como directorio de configuración a través de la propiedad CLAVE_SP2_CONFIG_REPOSITORY concatenando el directorio "Certificates"
		String defaultCertificatesPath = System.getProperty("user.dir") + "\\Certificates\\";
    	Boolean validDirectory = Boolean.FALSE;
    	File dirCertificates = new File(certificatesPath);

    	if (!dirCertificates.isDirectory()) {
    		logger.info("El directorio establecido en la propiedad 'certProxy2.certificatesPath' no existe o no es válido. Se utilizará el directorio de ejecución de la aplicación por defecto para la gestión de certificados. (" + defaultCertificatesPath + ")");
		} else {
			validDirectory = Boolean.TRUE;
		}
    	
    	if (Boolean.FALSE.equals(validDirectory)) {
    		certificatesPath = defaultCertificatesPath;
    		// Creamos directorio por defecto
    		new File(certificatesPath).mkdirs();
		}

		// Flag que indica si el daemon está o no activo
		if (daemonActivated == null || !"true".equals(daemonActivated)) {
			message = messagesProperties.getMessagesProperties().get("daemonCertificateProxy2.daemonCertificateProxy2Deactivated");
			logger.info(message);
			return;
		}

		message = messagesProperties.getMessagesProperties().get("daemonCertificateProxy2.daemonCertificateProxy2StartInfo");
		logger.info(message);

		try {
			InetAddress ip = InetAddress.getLocalHost();
			hostname = ip.getHostName();

			// Ejecucion del 'DaemonCertificateProxy2' en el nodo {hostname}
			message = messagesProperties.getMessageWithMultipleProps("daemonCertificateProxy2.actualNodeInfo", new ArrayList<String>(Arrays.asList(hostname)));
			logger.info(message);

			// Se abre una conexión HTTP o HTTPS de tipo GET que consume el servlet ofrecido por el modulo 'Proxy2' de CLAVE2.
			// La URL de esta conexión se configura en la propiedad 'certproxy2.endpoint' del fichero 'application.properties'.
			// A través de esta llamada obtenemos un XML con los diferentes certificados publicos existentes en la BDD de CLAVE. Se obtiene la siguiente informacion:
			// tipo de certificado: de firma (signing), o de cifrado (encryption)
			// nombre del certificado: 'proxy_sign_*' o 'proxy_cipher_*'
			// contenido del certificado en Base64
			// estado certificado: activo / inactivo

			message = messagesProperties.getMessageWithMultipleProps("daemonCertificateProxy2.connectingToURL", new ArrayList<String>(Arrays.asList(endpoint)));
			logger.info(message);

			HttpsURLConnection servletConnectionHttps = null;
			HttpURLConnection servletConnectionHttp = null;

			// Si la conexion es segura: 'https'
			if (Utils.isSecureConnection(endpoint)) {
				// Se carga el JKS 'TrustStore.jks'
				KeyStore truststore = KeyStore.getInstance(KeyStore.getDefaultType());
				char[ ] truststorePasswordChars = truststorePassword.toCharArray();
				try (FileInputStream truststoreFile = new FileInputStream(truststorePath)) {
					truststore.load(truststoreFile, truststorePasswordChars);
				}
				// Create and initialize the TrustManagerFactory with the truststore
				TrustManagerFactory trustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
				// Initialize the TrustManagerFactory with the loaded truststore
				trustManagerFactory.init(truststore);
				// Clear sensitive data from memory as soon as it's no longer needed
				Arrays.fill(truststorePasswordChars, '0');

				SSLContext sslContext = SSLContext.getInstance("TLSv1.2");
				sslContext.init(null, trustManagerFactory.getTrustManagers(), new SecureRandom());
				SSLSocketFactory socketFactory = sslContext.getSocketFactory();

				URL httpsUrl = new URL(endpoint);
				servletConnectionHttps = (HttpsURLConnection) httpsUrl.openConnection();
				servletConnectionHttps.setSSLSocketFactory(socketFactory);
				servletConnectionHttps.setRequestMethod("GET");
				servletConnectionHttps.setRequestProperty("Content-Type", "application/octet-stream");

			// Si la conexion no es segura: 'http'
			} else {
				URL httpUrl = new URL(endpoint);
				servletConnectionHttp = (HttpURLConnection) httpUrl.openConnection();
				servletConnectionHttp.setRequestMethod("GET");
				servletConnectionHttp.setRequestProperty("Content-Type", "application/octet-stream");
			}

			Map<String, String> certificatesStateMap = new HashMap<String, String>();
			Map<String, String> certificatesUpdatedMap = new HashMap<String, String>();

			// Se recuperan los pares atributo / valor del fichero 'certificates.properties'
			Properties certificatesProperties = Utils.loadCertificatesProperties(certificatesPath + "certificates.properties");
			Enumeration<Object> keys = certificatesProperties.keys();
			while (keys.hasMoreElements()) {
				String certificateName;
				String key = (String) keys.nextElement();
				String val = (String) certificatesProperties.get(key);

				if (key.endsWith(".state")) {
					certificateName = key.replace(".state", "");
					certificatesStateMap.put(certificateName, val);
				}
				if (key.endsWith(".updated")) {
					certificateName = key.replace(".updated", "");
					certificatesUpdatedMap.put(certificateName, val);
				}
			}

			// Una vez recuperados los pares atributo / valor del fichero 'certificates.properties', se eliminan para dejar el fichero de propiedades en blanco
			certificatesProperties.clear();

			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			// disable DOCTYPE declaration
			dbf.setFeature("http://apache.org/xml/features/disallow-doctype-decl", true);
			// Process XML securely, avoid attacks like XML External Entities (XXE)
			dbf.setFeature(XMLConstants.FEATURE_SECURE_PROCESSING, true);

			// Parse XML file
			DocumentBuilder db = dbf.newDocumentBuilder();
			Document doc;
			if (Utils.isSecureConnection(endpoint)) {
				doc = db.parse(servletConnectionHttps.getInputStream());
			} else {
				doc = db.parse(servletConnectionHttp.getInputStream());
			}
			doc.getDocumentElement().normalize();

			String certificateUse;
			String certificateKeyName;
			String certificateX509Certificate;
			String certificateMgmtData;

			NodeList IDPSSODescriptorNodeList = doc.getElementsByTagName("md:IDPSSODescriptor");
			Node IDPSSODescriptorNode = IDPSSODescriptorNodeList.item(0);

			for (int i = 0; i < IDPSSODescriptorNode.getChildNodes().getLength(); i++) {
				Node KeyDescriptorNode = IDPSSODescriptorNode.getChildNodes().item(i);
				if (KeyDescriptorNode.getNodeType() == Node.ELEMENT_NODE && KeyDescriptorNode.getNodeName() == "md:KeyDescriptor") {
					Element KeyDescriptorNodeElement = (Element) KeyDescriptorNode;

					// Tipo de certificado: de firma (signing) / de cifrado (encryption)
					certificateUse = KeyDescriptorNodeElement.getAttribute("use");

					for (int j = 0; j < KeyDescriptorNode.getChildNodes().getLength(); j++) {
						Node KeyInfoNode = KeyDescriptorNode.getChildNodes().item(j);
						if (KeyInfoNode.getNodeType() == Node.ELEMENT_NODE) {
							Element KeyInfoElement = (Element) KeyInfoNode;

							// Nombre de certificado: de firma (proxy_sign_*) / de cifrado (proxy_cipher_*)
							certificateKeyName = KeyInfoElement.getElementsByTagName("ds:KeyName").item(0).getTextContent();

							Node X509DataNode = KeyInfoElement.getElementsByTagName("ds:X509Data").item(0);
							Element X509DataElement = (Element) X509DataNode;
							// Valor del certificado en Base64
							certificateX509Certificate = X509DataElement.getElementsByTagName("ds:X509Certificate").item(0).getTextContent();

							// Estado del certificado: active / inactive
							certificateMgmtData = KeyInfoElement.getElementsByTagName("ds:MgmtData").item(0).getTextContent();

							// INICIO ALGORITMO DE ACTUZALIZACIÓN Y RECUPERACIÓN DE CERTIFICADOS PÚBLICOS
							// Si el nombre del certificado obtenido a través del XML existe en el fichero de propiedades 'certificates.properties'...
							if (certificatesStateMap.get(certificateKeyName) != null) {
								// Si el estado del certificado obtenido es igual al del fichero de propiedades...
								if (certificatesStateMap.get(certificateKeyName).equals(certificateMgmtData)) {
									if (certificateUse.equals("signing")) {
										// Certificado de firma {certificateKeyName} ya existe
										message = messagesProperties.getMessageWithMultipleProps("daemonCertificateProxy2.signatureCertificateExists", new ArrayList<String>(Arrays.asList(certificateKeyName)));
									} else {
										// Certificado de cifrado {certificateKeyName} ya existe
										message = messagesProperties.getMessageWithMultipleProps("daemonCertificateProxy2.encryptionCertificateExists", new ArrayList<String>(Arrays.asList(certificateKeyName)));
									}
									logger.info(message);

									// Si el estado del certificado obtenido es diferente al del fichero de propiedades...
								} else {
									// Se descarga el certificado en el path de certificados y se actualiza en el fichero de propiedades 'certificates.properties'
									downloadCertificate("certificateUpdated", certificateUse, certificateKeyName, certificateX509Certificate, certificateMgmtData, certificatesStateMap, certificatesUpdatedMap);
								}

								// Si el nombre del certificado obtenido a traves del XML no existe en el fichero de propiedades 'certificates.properties'...
							} else {
								// Se descarga el certificado en el path de certificados y se inserta en el fichero de propiedades 'certificates.properties'
								downloadCertificate("certificateDownloaded", certificateUse, certificateKeyName, certificateX509Certificate, certificateMgmtData, certificatesStateMap, certificatesUpdatedMap);
							}
						}
					}
				}
			}

			// Se rellena el fichero 'certificates.properties' con la nueva información de los certificados actualizados
			Iterator<Map.Entry<String, String>> iterator = certificatesStateMap.entrySet().iterator();
			while (iterator.hasNext()) {
				String certificateName;
				Map.Entry<String, String> pairs = iterator.next();
				certificateName = pairs.getKey();
				certificatesProperties.setProperty(certificateName.concat(".state"), pairs.getValue());
				certificatesProperties.setProperty(certificateName.concat(".updated"), certificatesUpdatedMap.get(certificateName));
				iterator.remove(); // avoids a ConcurrentModificationException
			}

			FileOutputStream out = new FileOutputStream(certificatesPath.concat("certificates.properties"));
			certificatesProperties.store(out, null);
			out.close();

		} catch (UnknownHostException e) {
			// No se ha podido obtener el 'hostname' del servidor
			message = messagesProperties.getMessagesProperties().get("daemonCertificateProxy2.unknownHostname");
			logger.error(message, e);
			return;

		} catch (Exception e) {
			logger.error("Error en DaemonCertificateProxy2.recoverAndUpdatePublicCertificates", e);
			return;
		}

		message = messagesProperties.getMessageWithMultipleProps("daemonCertificateProxy2.daemonCertificateProxy2EndInfo", new ArrayList<String>(Arrays.asList(certificatesPath)));
		logger.info(message);
	}

	/**
	 * Function to physically download the certificate obtained from the XML
	 *
	 * @param actionType
	 * @param certificateUse
	 * @param certificateKeyName
	 * @param certificateX509Certificate
	 * @param certificateMgmtData
	 * @param certificatesStateMap
	 * @param certificatesUpdatedMap
	 * @throws IOException
	 */
	private void downloadCertificate(String actionType, String certificateUse, String certificateKeyName, String certificateX509Certificate,
			String certificateMgmtData, Map<String, String> certificatesStateMap, Map<String, String> certificatesUpdatedMap) throws IOException {

		SimpleDateFormat isoDate = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ");
		String nowAsISO = isoDate.format(new Date());

		Utils.writeFile(Base64.getDecoder().decode(certificateX509Certificate), certificatesPath.concat(certificateKeyName).concat(".cer"));

		certificatesStateMap.put(certificateKeyName, certificateMgmtData);
		certificatesUpdatedMap.put(certificateKeyName, nowAsISO);

		String message;
		// Actualización de certificado ya existente
		if (actionType.equals("certificateUpdated")) {
			if (certificateUse.equals("signing")) {
				// Certificado de firma '{certificateKeyName}' actualizado
				message = messagesProperties.getMessageWithMultipleProps("daemonCertificateProxy2.signatureCertificateUpdated", new ArrayList<String>(Arrays.asList(certificateKeyName)));
			} else {
				// Certificado de cifrado '{certificateKeyName}' actualizado
				message = messagesProperties.getMessageWithMultipleProps("daemonCertificateProxy2.encryptionCertificateUpdated", new ArrayList<String>(Arrays.asList(certificateKeyName)));
			}

			// Descarga de certificado no existente
		} else {
			if (certificateUse.equals("signing")) {
				// Certificado de firma '{certificateKeyName}' descargado
				message = messagesProperties.getMessageWithMultipleProps("daemonCertificateProxy2.signatureCertificateDownloaded", new ArrayList<String>(Arrays.asList(certificateKeyName)));
			} else {
				// Certificado de cifrado '{certificateKeyName}' descargado
				message = messagesProperties.getMessageWithMultipleProps("daemonCertificateProxy2.encryptionCertificateDownloaded", new ArrayList<String>(Arrays.asList(certificateKeyName)));
			}

		}
		logger.info(message);
	}
}
