/* 
/*******************************************************************************
 * Copyright (C) 2024 Secretaría General de la Administración Digital, Gobierno de España
 * This program is licensed and may be used, modified and redistributed under the  terms
 * of the European Public License (EUPL), either version 1.1 or (at your option)
 * any later version as soon as they are approved by the European Commission.
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and
 * more details.
 * You should have received a copy of the EUPL1.1 license
 * along with this program; if not, you may find it at
 * http:joinup.ec.europa.eu/software/page/eupl/licence-eupl
 ******************************************************************************/

/** 
 * <b>File:</b><p>es.clave.SP2.RequestSAML.java.</p>
 * <b>Description:</b><p>Class with shared utilities within the Kit.</p>
 * <b>Project:</b><p>Citizen identification and authentication platform.</p>
 * <b>Date:</b><p>03/10/2024.</p>
 * @author Gobierno de España.
 * @version 1.0, 03/10/2024.
 */
package es.clave.SP2.util;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Properties;

import org.opensaml.core.xml.config.XMLObjectProviderRegistrySupport;
import org.opensaml.core.xml.io.MarshallingException;
import org.opensaml.saml.common.SAMLObjectContentReference;
import org.opensaml.security.SecurityException;
import org.opensaml.security.credential.Credential;
import org.opensaml.security.x509.BasicX509Credential;
import org.opensaml.xmlsec.EncryptionConfiguration;
import org.opensaml.xmlsec.SecurityConfigurationSupport;
import org.opensaml.xmlsec.keyinfo.KeyInfoGenerator;
import org.opensaml.xmlsec.keyinfo.KeyInfoGeneratorFactory;
import org.opensaml.xmlsec.keyinfo.KeyInfoGeneratorManager;
import org.opensaml.xmlsec.keyinfo.NamedKeyInfoGeneratorManager;
import org.opensaml.xmlsec.signature.KeyInfo;
import org.opensaml.xmlsec.signature.SignableXMLObject;
import org.opensaml.xmlsec.signature.Signature;
import org.opensaml.xmlsec.signature.support.ContentReference;
import org.opensaml.xmlsec.signature.support.SignatureConstants;
import org.opensaml.xmlsec.signature.support.SignatureException;
import org.opensaml.xmlsec.signature.support.Signer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

/** 
 * <p>Class with shared utilities within the Kit.</p>
 * <b>Project:</b><p>Citizen identification and authentication platform.</p>
 * @version 1.0, 03/10/2024.
 */
@Component
public final class Utils {

    private static final Logger logger = LoggerFactory.getLogger(Utils.class);

    public static final String HTTPS_PROTOCOL = "https";

    /**
     * Function to sign SAML requests going from Kit to Proxy2
     * 
     * @param <T> signableObject Object to sign
     * @param key Private key associated with the signing certificate
     * @param certificate Signing certificate
     * @param signatureAlgorithm Signature algorithm, set to 'signature.algorithm'
     * @throws Exception
     */
    public static <T extends SignableXMLObject> void signRequest(T signableObject, PrivateKey key, X509Certificate certificate, String signatureAlgorithm) throws Exception {

        try {
            BasicX509Credential credential = new BasicX509Credential(certificate, key);

            Signature signature = (Signature) XMLObjectProviderRegistrySupport.getBuilderFactory().getBuilder(Signature.DEFAULT_ELEMENT_NAME).buildObject(Signature.DEFAULT_ELEMENT_NAME);

            // Establecer el credential (la clave privada y el certificado X509)
            signature.setSigningCredential(credential);
            signature.setSignatureAlgorithm(signatureAlgorithm);
            signature.setKeyInfo(getKeyInfo(credential));
            signature.setCanonicalizationAlgorithm(SignatureConstants.ALGO_ID_C14N_EXCL_OMIT_COMMENTS);

            // Asignar la firma a la petición o el objeto firmable
            signableObject.setSignature(signature);

            String digestAlgorithm = "http://www.w3.org/2001/04/xmlenc#sha512";

            List<ContentReference> contentReferences = signature.getContentReferences();
            if (!contentReferences.isEmpty()) {
                ((SAMLObjectContentReference) contentReferences.get(0)).setDigestAlgorithm(digestAlgorithm);
            } else {
                logger.error("Unable to set DigestMethodAlgorithm - algorithm {} not set", digestAlgorithm);
            }

            logger.info("Marshall samlToken.");
            XMLObjectProviderRegistrySupport.getMarshallerFactory().getMarshaller(signableObject).marshall(signableObject);

            logger.info("Sign samlToken.");
            Signer.signObject(signature);

        } catch (MarshallingException e) {
            logger.error("signRequest - Error al realizar el marshalling.", e);
            throw e;

        } catch (SignatureException e) {
            logger.error("signRequest - Error al realizar la firma.", e);
            throw e;
        }
    }

    private static KeyInfo getKeyInfo(Credential credential) throws SecurityException {
        EncryptionConfiguration secConfiguration = SecurityConfigurationSupport.getGlobalEncryptionConfiguration();
        NamedKeyInfoGeneratorManager namedKeyInfoGeneratorManager = secConfiguration.getDataKeyInfoGeneratorManager();
        KeyInfoGeneratorManager keyInfoGeneratorManager = namedKeyInfoGeneratorManager.getDefaultManager();
        KeyInfoGeneratorFactory keyInfoGeneratorFactory = keyInfoGeneratorManager.getFactory(credential);
        KeyInfoGenerator keyInfoGenerator = keyInfoGeneratorFactory.newInstance();
        KeyInfo keyInfo = keyInfoGenerator.generate(credential);

        return keyInfo;
    }

    /**
     * Function that checks if a URL uses the secure protocol
     * 
     * @param urlService
     * @return true or false
     */
    public static boolean isSecureConnection(String urlService) {
        return urlService.toLowerCase().startsWith(HTTPS_PROTOCOL);
    }

    /**
     * Function that retrieves properties from a properties file
     * 
     * @param propertiesFilePath Path and name of the properties file
     * @return Properties
     * @throws Exception
     */
    public static Properties loadCertificatesProperties(String propertiesFilePath) throws Exception {
        Properties properties = new Properties();

        try {
            File f = new File(propertiesFilePath);
            try (FileReader fileReader = new FileReader(f)) {
                properties.load(fileReader);
            }

        } catch (FileNotFoundException fnfe) {
            logger.error("El fichero '" + propertiesFilePath + "' no existe en la ruta '" + propertiesFilePath + "'. Se creará un fichero nuevo.");

        }

        return properties;
    }

    /**
     * Function to write data from a byte array to a file
     * 
     * @param data Array of data bytes
     * @param filename Path and name of the destination file
     * @throws IOException
     */
    public static void writeFile(byte[ ] data, String filename) throws IOException {
        try (ByteArrayInputStream bais = new ByteArrayInputStream(data);
                FileOutputStream fos = new FileOutputStream(new File(filename))) {

            byte[ ] buffer = new byte[1024];
            int bytesReaded;
            while ((bytesReaded = bais.read(buffer)) >= 0) {
                fos.write(buffer, 0, bytesReaded);
            }

        } catch (IOException e) {
            logger.error("Error writing file", e);
            throw e;
        }
    }
}
